from SSUphysicsTools.getting_data import get_all_csv_paths_in_data, read_csv_Tektronix
import numpy as np
from pandas import DataFrame
from typing import Optional
import matplotlib.pyplot as plt

class DataPlots(get_all_csv_paths_in_data):
    def __init__(self,plot_name:tuple[str], flatten:Optional[bool]=False):
        '''
        :param plot_name: tuple[str], the name of the plots.
        '''
        super().__init__(flatten=flatten)
        self.plot_name = plot_name
    
    # plots
    def data_plots(self, is_save:Optional[bool]=False,dpi:Optional[int]=300):
        '''
        Automaticaly plots the original data generated by Tektronix.
        :param is_save: bool, save the plots or not. no support for flatten=True.
        > warning: no support for flatten=True yet.
        '''
        plot_name=self.plot_name
        flatten=self.flatten
        all_csv_list=self.all_csv_list

        if flatten:
            for csv in all_csv_list:
                results = read_csv_Tektronix(csv)
                data=results['data']
                plt.plot(data['Time'], data['Voltage'])
                plt.xlabel('Time (s)')
                plt.ylabel('Voltage (V)')
                plt.show()
                plt.close()

        # flatten=False
        all_csv_list=all_csv_list
        assert len(plot_name)==len(all_csv_list), 'The number of plot names should be the same as the number of csv files.'

        for i,csv_list in enumerate(all_csv_list):
            results = np.array([read_csv_Tektronix(csv_list[0]),read_csv_Tektronix(csv_list[1])])
            data1=results[0]['data']
            data2=results[1]['data']
            plt.plot(data1['Time'], data1['Voltage'])
            plt.plot(data2['Time'], data2['Voltage'])
            plt.xlabel('Time (s)')
            plt.ylabel('Voltage (V)')
            plt.legend(['CH1', 'CH2'])
            plt.title(f'{plot_name[i]}')
            for ii in (0,):
                plt.axhline(y=ii, color='gray', zorder=-1)
            plt.axvline(x=0, color='gray', zorder=-1)
            if is_save:
                print(f'Saving {plot_name[i]}')
                plt.savefig(f'fig/{plot_name[i]}.png',dpi=dpi)
            plt.show()
            plt.close()
    
    def Lissajous_plots(self):
        '''
        Tektronix oscilloscope에서 생성된 데이터를 Lissajous plot으로 시각화합니다.
        '''
        plot_name=self.plot_name
        all_csv_list=self.all_csv_list

        for i,csv_list in enumerate(all_csv_list):
            results = np.array([read_csv_Tektronix(csv_list[0]),read_csv_Tektronix(csv_list[1])])
            data1=results[0]['data']
            data2=results[1]['data']
            plt.plot(data1['Voltage'], data2['Voltage'])
            plt.xlabel('CH1 Voltage (V)')
            plt.ylabel('CH2 Voltage (V)')
            plt.title(f'{plot_name[i]}')
            plt.show()
            plt.close()
    
    def bode_plots(self):
        assert self.flatten==False, 'The flatten should be False to use bode_plots method.'
        plot_name=self.plot_name
        all_csv_list=self.all_csv_list

        for i,csv_list in enumerate(all_csv_list):
            results = np.array([read_csv_Tektronix(csv_list[0]),read_csv_Tektronix(csv_list[1])])
            data1=results[0]['data']
            data2=results[1]['data']
            plt.plot(data1['Time'], data1['Voltage'])
            plt.plot(data2['Time'], data2['Voltage'])
            plt.xlabel('Time (s)')
            plt.ylabel('Voltage (V)')
            plt.legend(['CH1', 'CH2'])
            plt.title(f'{plot_name[i]}')
            for i in (-1,0,1):
                plt.axhline(y=i, color='gray', zorder=-1)
            plt.axvline(x=0, color='gray', zorder=-1)
            plt.show()
            plt.close()

def plot_table(data:DataFrame, title:str, table_font_size:Optional[int]=8,save_dir:Optional[str]=None ,is_save:Optional[bool]=False)->None:
    """
    데이터프레임을 테이블 형태로 시각화합니다.
    :param data: 데이터프레임
    :param title: 테이블 제목
    :param save_dir: 저장 경로, is_save가 True일 때만 사용, None일 때는 project root directory에 저장
    :param is_save: 저장 여부
    :return: None
    """
    # data의 모든 값을 문자열로 변환
    data=data.astype(str)

    # 그림과 축 생성
    fig, ax = plt.subplots()

    # 축을 숨김
    ax.axis('tight')
    ax.axis('off')

    # 테이블 생성
    table = ax.table(cellText=data.values, colLabels=data.columns, cellLoc='center', loc='center', bbox=[0, 0, 1, 1])

    # 헤더 셀 음영 처리
    for key, cell in table.get_celld().items():
        if key[0] == 0:  # 헤더 행
            cell.set_facecolor('#D3D3D3')  # 연한 회색
            cell.set_text_props(weight='bold')  # 텍스트 굵게 설정
        

    table.set_fontsize(table_font_size)  # 테이블 전체 폰트 크기 설정
    table.auto_set_font_size(False)  # 폰트 크기 자동 조정 비활성화
    table.auto_set_column_width(col=list(range(len(data.columns))))  # 셀 너비 자동 조정 비활성화
    title_font = {'fontsize': 16, 'fontweight': 'bold', 'fontname': 'Times New Roman','color': 'black'}
    ax.set_title(title,fontdict=title_font)
    plt.tight_layout()
    plt.show()
    if is_save:
        if save_dir is None:
            save_dir = './'
            plt.savefig(f'{title}.png', bbox_inches='tight', dpi=300)

        else: # 저장하지 않을 경우 그림 닫기
            plt.savefig(f'{save_dir}/{title}.png', bbox_inches='tight', dpi=300)

    plt.close()


from ast import Global
import pandas as pd
from pandas import DataFrame
from typing import Optional
import os
import re
from collections import defaultdict
import numpy as np
import matplotlib.pyplot as plt

def pairwise_to_2d_list(flat_list:list)->list[any]:
    # 2개씩 묶어서 2차원 배열로 변환
    return [flat_list[i:i + 2] for i in range(0, len(flat_list), 2)]

def get_channel_csv_files(directory) -> dict[str,str]:
    """
    주어진 디렉토리 내의 CSV 파일 중 파일 이름에 'CH1', 'CH2' 등이 포함된 파일을 채널별로 딕셔너리로 반환합니다.
    
    :param directory: str, 디렉토리 경로
    :return: dict, 채널별 CSV 파일 경로 딕셔너리

    > Tektronix는 CSV 파일 이름에 'CH1', 'CH2' 등 채널 정보가 포함되어 있습니다. 또한 확장자가 csv가 아닌 CSV로 되어 있으니 주의하세요.
    """
    csv_files = [file for file in os.listdir(directory) if file.endswith('.CSV')] # csv가 아니라 CSV로 되어 있음.

    assert len(csv_files) > 0, 'No CSV files in the directory.'

    channel_files:defaultdict = defaultdict(list)
    
    for file in csv_files:
        match = re.search(r'CH\d+', file)
        if match:
            channel = match.group()
            channel_files[channel].append(os.path.join(directory, file))
    
    return dict(channel_files)

# Read the data from the CSV file
def read_csv_Tektronix(file_path:str,columns_name:Optional[str]=['Time', 'Voltage'], exclude_columns:Optional[list]=[0,1,2,5], metadata_lows:Optional[int]=18, name:Optional[str]=None) -> dict[str, DataFrame]:
    '''
    Read Tektronix's CSV file and return the data and metadata.
    :param file_path: str, the path of the CSV file.
    :param exclude_columns: list, the index of the columns to be excluded.
    :param metadata_lows: int, the number of rows to be extracted as metadata.
    :return: dict, the data and metadata.
    '''
    # CSV 파일에서 데이터 읽기
    data = pd.read_csv(file_path)
    
    # extract metadata
    metadata = data.iloc[:metadata_lows]

    # 제외할 열 이름 계산
    exclude_column_names = [data.columns[i] for i in exclude_columns]
    
    # 제외할 열 이름을 제외한 나머지 열 이름 선택
    usecols = [column for column in data.columns if column not in exclude_column_names]
    
    # CSV 파일 다시 읽기, 특정 열 제외
    data = pd.read_csv(file_path, usecols=usecols)

    # set the name of the data
    if name is not None:
        data.name=name

    # data 열 이름 지정
    if len(columns_name) == len(data.columns):
        data.columns = columns_name
    else:
        raise ValueError('The number of columns_name is not same as the number of columns in the data.: columns_name={}, data.columns={}'.format(len(columns_name), len(data.columns)))

    # results
    result = {
        'metadata': metadata,
        'data': data
    }
    return result

class get_all_subdirs_in_dir:
    def __init__(self, directory:Optional[str]='data'):
        '''
        주어진 디렉토리 내의 모든 폴더 이름을 가져옵니다.
        :param directory: Optional[str], the directory name
        '''
        self.directory=directory

    def get_sorted_folders_dir_by_number(self) -> list[str]:
        """
        주어진 디렉토리 내의 모든 폴더 이름에서 숫자를 추출하여, 숫자 기준으로 오름차순 정렬된 폴더 이름 목록을 반환합니다.
        :return: list, 숫자 기준으로 정렬된 폴더 이름 목록
        """
        directory=self.directory

        folder_names:list[str] = [name for name in os.listdir(directory) if os.path.isdir(os.path.join(directory, name))]
        
        # 폴더 이름에서 숫자를 추출하고, 숫자가 없는 경우 0으로 간주
        folder_numbers:list[str] = [(folder, int(re.search(r'\d+', folder).group()) if re.search(r'\d+', folder) else 0) for folder in folder_names]
        
        sorted_folders:list = sorted(folder_numbers, key=lambda x: x[1]) # 숫자 기준으로 폴더 이름 정렬
        return [directory+'/'+folder[0] for folder in sorted_folders]
    
    def get_all_directories(self) -> list[str]:
        """
        주어진 디렉토리 내의 모든 sub디렉토리 경로를 출력합니다.
        :param directory: str, 디렉토리 경로
        """
        directory=self.directory
        items = os.listdir(directory) # data directory는 package가 init되면서 생성됨.
        results=[]
        # 각 항목이 디렉토리인지 확인하고, 디렉토리인 항목의 이름을 출력
        for item in items:
            if os.path.isdir(os.path.join(directory, item)):
                results.append(f'{directory}/{item}')
        return results

class get_all_csv_paths_in_data(get_all_subdirs_in_dir):
    def __init__(self, flatten:Optional[bool]=False, custom_dir_name:Optional[bool]=False):
        '''
        data 폴더 내의 모든 csv 파일을 가져옵니다.
        :param flatten: bool, 만약 True이면, 실험 번호대로 시험 데이터를 묶는 게 아닌(2차원 배열), 1차원 배열로 변환합니다.
        :param custom_dir_name: Optional[list[str]], 만약 Tektronix의 오실로스코프가 지정한 실험 이름이 아닌 custom으로 실험 이름을 지정했다면 True로 설정합니다. False이면, 모든 실험 순서대로 데이터를 가져옵니다.
        '''
        self.directory = 'data'
        super().__init__(directory=self.directory)
        self.flatten = flatten
        self.custom_dir_name = custom_dir_name
        self.all_csv_list = self.__call__()

    def __rough_get_all_csv_paths(self, experi_dirs:list[str])->np.ndarray:
        '''
        flatten이나 custom_dir_name을 고려하지 않고, 모든 csv 파일을 가져옵니다.

        > 코드 가독성을 위해 본 함수를 만들었기에 private 함수로 만들었습니다.
        '''
        rough_all_csv_list = []
        # get all the csv files in the directory
        for dir in experi_dirs:
            temp_list=[get_channel_csv_files(dir)['CH1'][0],get_channel_csv_files(dir)['CH2'][0]]
            csv_list=temp_list
            rough_all_csv_list.append(csv_list)
        rough_all_csv_list:np.ndarray=np.array(rough_all_csv_list)
        return rough_all_csv_list

    def __call__(self)->np.ndarray:
        '''
        주어진 조건에 따라 모든 csv 파일을 가져옵니다.
        :return: np.ndarray, the list of the csv files.
        '''
        directory:str=self.directory
        flatten:bool=self.flatten
        custom_dir_name:bool=self.custom_dir_name
        experi_dirs= self.get_sorted_folders_dir_by_number() # Tektronix가 지정한 실험 이름을 사용.
        if custom_dir_name:
            experi_dirs=self.get_all_directories() # custom 실험 이름을 사용.

        all_csv_list=self.__rough_get_all_csv_paths(experi_dirs)

        # flatten
        if flatten:
            all_csv_list=all_csv_list.flatten()
        return all_csv_list


# exceptions.py
class DataDirectoryEmptyError(Exception):
    """'data' 디렉토리가 비어 있을 때 발생하는 예외"""
    def __init__(self, message="오류: 'data' 디렉토리에 파일이 없습니다. 실험 데이터를 'data' 디렉토리에 넣어주세요."):
        self.message = message
        super().__init__(self.message)


from scipy.optimize import curve_fit
import numpy as np
from typing import Callable, Optional
from pandas import DataFrame

# regression
class Regression:
    @staticmethod
    def cosine_regression(x_data:np.ndarray, y_data:np.ndarray, freq:Optional[str]=None)->dict:
        pure_model=lambda x,A,f,phi,offset: A*np.cos(2*np.pi*f*x+phi)+offset

        if freq: # freq is given
            freq_model:Callable=lambda x,A,phi,offset: pure_model(x,A,freq,phi,offset)
            freq_parms, _ = curve_fit(freq_model, x_data, y_data)
            fitted_func:Callable= lambda x: freq_model(x, *freq_parms)
            result:dict={'fitted_func':fitted_func,'parms':freq_parms}
            return result
        
        # freq is not given
        parms, _ = curve_fit(pure_model, x_data, y_data, maxfev=10000)
        fitted_func:callable= lambda x: pure_model(x, *parms)
        result:dict={'fitted_func':fitted_func,'parms':parms}
        return result

class Delate_offset(Regression):
    '''
    Regression을 사용해 offset을 데이터에서 제거합니다.
    '''
    def __init__(self, data:DataFrame, data_column:Optional[list[str]]=['Time','Voltage'],is_use_regression:Optional[bool]=False):
        '''
        Delate the offset from the data.
        :param data: DataFrame
        :param data_column: list[str], the column names of the data.
        :param is_use_regression: bool, if True, use the regression to find the offset. If False, use the mean value of the data.
        '''
        assert type(data) is DataFrame, 'The type of the data should be Dataframe. Not {}'.format(type(data))

        self.data=data
        self.is_use_regression=is_use_regression
        self.column_x:str=data_column[0]
        self.column_y:str=data_column[1]
        self.data_x=data[self.column_x]
        self.data_y=data[self.column_y]
        if self.is_use_regression:
            self.offset:float=self.cosine_regression(self.data_x,self.data_y)['parms'][3] # offset 값 찾기
        else:
            self.offset:float=self.data_y.mean()
    
    def __call__(self)->DataFrame:
        '''
        Delate the offset from the data.
        '''
        offset=self.offset
        data_x=self.data_x
        data_y=self.data_y
        data_y=data_y-offset
        results={self.column_x:data_x,self.column_y:data_y}
        return DataFrame(results)


# 주요 모듈 임포트
from .data_analysis import *
from .data_processing import *
from .getting_data import *
from .Plots import *
from .exceptions import DataDirectoryEmptyError

# 필요한 디렉토리가 있는 지 확인 후 생성
directory_list = ['fig','data']
for directory in directory_list:
    if not os.path.exists(directory):
        os.makedirs(directory)

# 'data' 디렉토리에 파일이 없는 경우 경고 문구 출력
data_directory = directory_list[1]
if os.path.exists(data_directory) and not os.listdir(data_directory):
    raise DataDirectoryEmptyError()


import numpy as np
from typing import Optional
from pandas import Series

class Phase_shift:
    def __init__(self, ch1_data:Series|np.ndarray, ch2_data:Series|np.ndarray):
        '''
        calculate the phase shift between ch1 and ch2.
        :param ch1_data: Series|np.ndarray, the first data.
        :param ch2_data: Series|np.ndarray, the second data.
        '''
        assert type(ch1_data) is Series or type(ch1_data) is np.ndarray, 'The type of the ch1_data should be Series or np.ndarray. Not {}'.format(type(ch1_data))

        self.ch1_data=ch1_data
        self.ch2_data=ch2_data
        self.ym_y0:dict=self.find_ym_y0()

    def find_min_x_vector(self,vectors:np.ndarray[:,2])->Series:
        '''
        2차원 벡터 리스트에서 x값의 크기가 가장 작은 벡터를 찾는다.
        '''
        # 벡터 리스트를 x값 기준으로 정렬
        abs_vector_x=np.abs(vectors[:,0]) # x값의 절댓값을 구한다.
        sorted_index = np.argsort(abs_vector_x) # 입력받은 벡터를 절댓값으로 만들고 크기가 작은 순서대로 정렬 이후 원래 인덱스를 반환
        minimal_index=sorted_index[0] # x값의 크기가 가장 작은 벡터의 인덱스
        smallest_vector = vectors[minimal_index] # x값의 크기가 가장 작은 벡터 반환
        # x값의 크기가 가장 작은 벡터 반환
        return smallest_vector

    # find ym and y0
    def find_ym_y0(self)->dict[str, float]:
        '''
        Find ym and y0
        :param data1: Series, the first data.
        :param data2: Series, the second data.
        :return: tuple, the ym and y0.
        '''
        ch1_data=self.ch1_data
        ch2_data=self.ch2_data

        # chanage the data type to numpy array
        if type(ch1_data) is Series:
            ch1_data=ch1_data.values
        if type(ch2_data) is Series:
            ch2_data=ch2_data.values
        
        assert len(ch1_data)==len(ch2_data), 'The length of the data should be the same.'

        # find y0
        ## ch2가 양수일 때의 ch1값이 0일때와 ch2가 음수일 때의 ch1값이 0일 때의 ch2값을 각각 나눠서 찾아야 한다.
        combined_data:np.ndarray[:,2]=np.column_stack((ch1_data,ch2_data))
        filtered_data1:np.ndarray[:,2]=combined_data[combined_data[:,1]>0] ## ch2가 양수일 때의 데이터
        filtered_data2:np.ndarray[:,2]=combined_data[combined_data[:,1]<0] ## ch2가 음수일 때의 데이터

        assert len(filtered_data1)>0 or len(filtered_data2)>0, 'The ch2 data should have positive or negative values.'

        ## ch1가 0에 가장 가까울 때의 ch2값을 찾는다.
        upper_y0=self.find_min_x_vector(filtered_data1)
        lower_y0=self.find_min_x_vector(filtered_data2)

        ## y0 계산, y0_source 계산
        y0=np.abs(upper_y0[1])+np.abs(lower_y0[1])
        y0_source=np.array([upper_y0,lower_y0])

        # find ym: ym은 ch2의 최대값과 최솟값의 차이다.
        sorted_indices_ch2=np.argsort(ch2_data)
        max_index_ch2=sorted_indices_ch2[-1]
        min_index_ch2=sorted_indices_ch2[0]
        ym=np.abs(ch2_data.max())+np.abs(ch2_data.min())
        ym_source=np.array([[ch1_data[max_index_ch2],ch2_data.max()],
                            [ch1_data[min_index_ch2],ch2_data.min()]
        ])
        
        return {'ym':ym, 'y0':y0,'ym_source':ym_source,'y0_source':y0_source}

    def __call__(self):
        '''
        Find the phase shift between ch1 and ch2.
        '''
        ym_y0=self.ym_y0
        ym=ym_y0['ym']
        y0=ym_y0['y0']
        
        # find the phase shift
        theta=np.arcsin((y0/ym))
        deg_theta=np.rad2deg(theta)
        if deg_theta<0: # 0~360도로 변환
            deg_theta=360+deg_theta
            
        if 0<deg_theta<90 or 270<deg_theta<360:
            return theta # theta is phase shift
        else:
            phase_shift=np.pi-theta
            return phase_shift

def get_error_rate(theory:np.ndarray, experiment:np.ndarray, absolute:Optional[bool]=True) -> np.ndarray:
    '''
    오차 백분율을 계산합니다.
    :param theory: np.ndarray, 이론값
    :param experiment: np.ndarray, 실험값
    :param absolute: Optional[bool]=True, True이면 절대 오차 백분율을 반환합니다.
    :return: np.ndarray, 오차 백분율

    오차 백분율은 다음과 같이 계산됩니다.
    $$\text{Error Rate} = |\frac{\text{Theory} - \text{Experiment}}{\text{Theory}}| \times 100$$
    이 때 absolute가 True이면 절대 오차 백분율을 반환하고, False이면 위의 식에서 절댓값이 제거된 체로 오차 백분율을 반환합니다.

    Tip: You can also use skit-learn's mean_squared_error function. See https://scikit-learn.org/stable/modules/generated/sklearn.metrics.mean_squared_error.html
    '''
    if absolute:
        return 100*np.abs(theory - experiment) / np.abs(theory)
    else:
        return 100*(theory - experiment) / np.abs(theory)


from setuptools import setup, find_packages

setup(
    name="SSUphysicsTools",
    version="2.1.1",
    packages=find_packages(),
    install_requires=[
        "numpy",
        "scipy",
        "matplotlib",
        "pandas"
    ],
    author="Taeyoung Kim",
    author_email="rocketman@soongsil.ac.kr",
    description="숭실대학교 물리학과 실험에 사용되는 데이터 처리 및 시각화 도구",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://github.com/kty1004/SSUphysicsExperiment_Supporting_tools?tab=readme-ov-file#readme",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.13.0',
    test_require=[
        "pytest"
    ],
    test_suite='tests',
)